'use strict';
const tencentcloud = require('tencentcloud-sdk-nodejs');
const Abstract     = require('./abstract');
const utils        = require('./utils');
const assert       = require('assert');
const fs           = require('fs');
const _            = require('lodash');
const filesize     = require('filesize');
const request      = require('request');
const util         = require('util');

const models       = tencentcloud.scf.v20180416.Models;
const ScfZipCodeSizeLimit = 20 * 1024 * 1024;
const ScfUploadSliceLimit = 8 * 1024 * 1024;

class DeployFunction extends Abstract {
    constructor(appid, secret_id, secret_key, options) {
        super(appid, secret_id, secret_key, options);
    }

    async deploy(ns, funcObject, packagePath) {
        const func = await this.getFunction(ns, funcObject.FuncName);
        if (!func)
           await this.createFunction(ns, funcObject, packagePath);
        else {
           await this.updateFunctionCode(ns, funcObject, packagePath);
           return func;
        }
        return null;
    }

    async updateFunctionCode(ns, funcObject, packagePath) {
        const updateFuncRequest = {
            Region: funcObject.Properties.Region,
            FunctionName: funcObject.FuncName,
            Handler: funcObject.Properties.Handler,
            Namespace: ns,
            CosBucketName: funcObject.Properties.CodeUri.Bucket,
            CosObjectName: funcObject.Properties.CodeUri.Key
        };

        const handler = util.promisify(this.scfClient.UpdateFunctionCode.bind(this.scfClient));
        try {
            this.logger('Updating function code from cos');
            const result = await handler(updateFuncRequest);
            this.logger('Request id', result.RequestId);
            return result;
        } catch (e) {
            throw e;
        }
    }

    async createFunction(ns, funcObject, packagePath) {
        let funcCode;

        if (funcObject.Properties.CodeUri.Key) {
            funcCode = {
                CosBucketName: funcObject.Properties.CodeUri.Bucket, 
                CosObjectName: funcObject.Properties.CodeUri.Key
            };
        } else {
            const size = fs.statSync(packagePath).size;
            assert(size <= ScfZipCodeSizeLimit, `file size(${filesize(size)}) exceeds maximum(${filesize(ScfZipCodeSizeLimit)}) limit.`);

            const zipBinContent = fs.readFileSync(packagePath);
            funcCode = { 
                ZipFile: zipBinContent.toString('base64') 
            };
        }

        const createFuncRequest = {
            Region: funcObject.Properties.Region,
            FunctionName: funcObject.FuncName,
            Code: funcCode,
            Namespace: ns,
            Runtime: funcObject.Properties.Runtime
        };

        if (!_.isEmpty(funcObject.Properties.Environment)) {
            const env = funcObject.Properties.Environment;
            
            createFuncRequest.Environment = {
                Variables: []
            };
            for (let key in env.Variables) {
                const item = {
                    Key: key,
                    Value: env.Variables[key].toString()
                };
                createFuncRequest.Environment.Variables.push(item);
            }
        }
        const handler = util.promisify(this.scfClient.CreateFunction.bind(this.scfClient));
        try {
            this.logger('Uploading function', packagePath);
            const result = await handler(createFuncRequest);
            this.logger('Request id', result.RequestId);
            return result;
        } catch (e) {
            throw e;
        }
    }

    async getFunction(ns, funcName, showCode) {
        const getFuncRequest = new models.GetFunctionRequest();
        getFuncRequest.FunctionName  = funcName;
        getFuncRequest.Namespace     = ns;
        getFuncRequest.ShowCode      = showCode ? 'TRUE' : 'FALSE';

        const handler = util.promisify(this.scfClient.GetFunction.bind(this.scfClient));
        try {
            return await handler(getFuncRequest)
        } catch (e) {
            if (e.code == 'ResourceNotFound.FunctionName' || e.code == 'ResourceNotFound.Function')
                return null;
            throw e;
        }
    }

    async updateConfiguration(ns, oldFunc, funcObject) {
        const configArgs = {};
        if (!oldFunc) {
            if (funcObject.Properties.Description)
                configArgs.Description = funcObject.Properties.Description;
            if (funcObject.Properties.Runtime)
                configArgs.Runtime = funcObject.Properties.Runtime;
            if (funcObject.Properties.Role)
                configArgs.Role = funcObject.Properties.Role;
            if (funcObject.Properties.ClsLogsetId)
                configArgs.ClsLogsetId = funcObject.Properties.ClsLogsetId;
            if (funcObject.Properties.ClsTopicId)
                configArgs.ClsTopicId = funcObject.Properties.ClsTopicId;
            if (funcObject.Properties.VpcConfig) 
                configArgs.VpcConfig = funcObject.Properties.VpcConfig;

        } else {
            if (funcObject.Properties.Description && oldFunc.Description != funcObject.Properties.Description)
                configArgs.Description = funcObject.Properties.Description;
            if (funcObject.Properties.Runtime && oldFunc.Runtime != funcObject.Properties.Runtime)
                configArgs.Runtime = funcObject.Properties.Runtime;
            if (funcObject.Properties.Role && oldFunc.Role != funcObject.Properties.Role)
                configArgs.Role = funcObject.Properties.Role;
            if (funcObject.Properties.ClsLogsetId && oldFunc.ClsLogsetId != funcObject.Properties.ClsLogsetId)
                configArgs.ClsLogsetId = funcObject.Properties.ClsLogsetId;
            if (funcObject.Properties.ClsTopicId && oldFunc.ClsTopicId != funcObject.Properties.ClsTopicId)
                configArgs.ClsTopicId = funcObject.Properties.ClsTopicId;
            if (!_.isEmpty(funcObject.Properties.VpcConfig) && 
                (oldFunc.VpcConfig.VpcId != funcObject.Properties.VpcConfig.VpcId || 
                oldFunc.VpcConfig.SubnetId != funcObject.Properties.VpcConfig.SubnetId)) 
                configArgs.VpcConfig = funcObject.Properties.VpcConfig;
            if (oldFunc.Timeout != funcObject.Properties.Timeout)
                configArgs.Timeout = funcObject.Properties.Timeout;
            if (oldFunc.MemorySize != funcObject.Properties.MemorySize)
                configArgs.MemorySize = funcObject.Properties.MemorySize;
        }

        const envs = [];
        const oldEnvs = [];
        if (oldFunc && oldFunc.Environment && oldFunc.Environment.Variables && 
            funcObject.Properties.Environment && _.isObject(funcObject.Properties.Environment.Variables)) {
            for (let key in funcObject.Properties.Environment.Variables) {
                const value = funcObject.Properties.Environment.Variables[key];
                const oldItem = _.find(oldFunc.Environment.Variables, (data) => {
                    if (data.Key == key && data.Value == value)
                        return data;
                });
                if (!_.isEmpty(oldItem)) {
                    oldEnvs.push(oldItem);
                    continue;
                }

                const item = {
                    Key: key,
                    Value: value
                }
                envs.push(item);
            }
        }

        if (!_.isEmpty(envs)) {
            configArgs.Environment = {
                Variables: [...envs, ...oldEnvs]
            };
        } else
        if (oldFunc && oldFunc.Environment && _.size(oldEnvs) != _.size(oldFunc.Environment.Variables)) {
            configArgs.Environment = {
                Variables: oldEnvs
            };
        } 

        // delete all evn
        if (oldFunc && oldFunc.Environment && _.size(oldFunc.Environment.Variables) > 0 &&
                configArgs.Environment && _.size(configArgs.Environment.Variables) == 0) {
            configArgs.Environment = {
                Variables: [{
                    Key: '',
                    Value: ''
                }]
            };
        }

        if (!_.isEmpty(configArgs)) {
            configArgs.Region = this.options.region;
            configArgs.FunctionName = funcObject.FuncName;
            configArgs.Namespace = ns;
            if (!configArgs.Timeout)
                configArgs.Timeout = funcObject.Properties.Timeout;
            if (!configArgs.MemorySize)
                configArgs.MemorySize = funcObject.Properties.MemorySize;
            
            const handler = util.promisify(this.scfClient.UpdateFunctionConfiguration.bind(this.scfClient));
            try {
                this.logger(`Updating function ${funcObject.FuncName} configure ${JSON.stringify(configArgs)}`);
                const result = await handler(configArgs);
                this.logger('Request id', result.RequestId);
            } catch (e) {
                throw e;    
            }
        } else 
            this.logger('configure not changed. function %s', funcObject.FuncName);
   
    }

    async uploadPackage2Cos(bucketName, key, filePath) {
        
        const region = this.options.region;
        const cosBucketNameFull = util.format('%s-%s', bucketName, this.appid);

        const getService      = util.promisify(this.cosClient.getService.bind(this.cosClient));
        const putBucket       = util.promisify(this.cosClient.putBucket.bind(this.cosClient));
        const putObject       = util.promisify(this.cosClient.putObject.bind(this.cosClient));
        const sliceUploadFile = util.promisify(this.cosClient.sliceUploadFile.bind(this.cosClient));
        try {
            // get region all bucket list
            const buckets = await getService({Region: region});
            const findBucket = _.find(buckets.Buckets, (item) => {
                if (item.Name == cosBucketNameFull)
                    return item;
            });

            let result;
            // create a new bucket
            if (_.isEmpty(findBucket)) {
                const putArgs = {
                    Bucket: cosBucketNameFull,
                    Region: region
                }
                await putBucket(putArgs);
            }

            if (fs.statSync(filePath).size <= ScfUploadSliceLimit) {
                const objArgs = {
                    Bucket: cosBucketNameFull, 
                    Region: region,
                    Key: key,
                    Body: fs.createReadStream(filePath),
                    ContentLength: fs.statSync(filePath).size
                };
                result = await putObject(objArgs);
            } else {
                const sliceUploadRequest = {
                    Bucket: cosBucketNameFull,
                    Region: region,
                    Key: key,
                    FilePath: filePath,
                    onTaskReady: function(taskId) {
                    },
                    onProgress: function (progressData) {
                    }
                };

                result = await sliceUploadFile(sliceUploadRequest);
            }

            this.logger(`Uploaded file[${result.ETag}] success. assess url ${result.Location}`);
        } catch (e) {
            throw e;    
        }

        return {
            CosBucketName: bucketName,
            CosObjectName: '/' + key
        };
    }

    async uploadService2Cos(bucketName, key, object) {
        let value;
        if (_.isObject(object))
            value = JSON.stringify(object);
        else 
            value = object;

        const region = this.options.region;
        const cosBucketNameFull = util.format('%s-%s', bucketName, this.appid);

        const putObjectRequest = {
            Bucket: cosBucketNameFull, 
            Region: region,
            Key: key,
            Body: value,
        };
        const putObject = util.promisify(this.cosClient.putObject.bind(this.cosClient));
        try {
            return await putObject(putObjectRequest);
        } catch (e) {
            throw e;
        }
    }

    async createTags(ns, funcName, tags) {
        if (_.isEmpty(tags)) return;
        const func = await this.getFunction(ns, funcName);
        if (!func) 
            throw new Error(`Function ${funcName} dont't exists`);

        const resource = util.format('qcs::scf:%s::lam/%s', this.options.region, func.FunctionId);
        const modifyTagsRequest = {
            Resource: resource,
            ReplaceTags: [],
            DeleteTags: []
        };

        const getTagsRequest = {
            ResourceRegion: this.options.region,
            ResourceIds: [func.FunctionId],
            ResourcePrefix: 'lam',
            ServiceType: 'scf',
            Limit: 1000
        }

        const getResourceTags    = util.promisify(this.tagClient.DescribeResourceTagsByResourceIds.bind(this.tagClient));
        const modifyResourceTags = util.promisify(this.tagClient.ModifyResourceTags.bind(this.tagClient));
        try {
            const result = await getResourceTags(getTagsRequest);

            const len = _.size(result.Tags);
            for (let i = 0; i < len; i++) {
                const oldTag = result.Tags[i];
                const ret = _.find(tags, (value, key) => {
                    if (oldTag.TagKey == key)
                        return true;
                });
                if (!ret) {
                    modifyTagsRequest.DeleteTags.push({
                        TagKey: oldTag.TagKey
                    });
                }
            }

            for (let key in tags) {
                modifyTagsRequest.ReplaceTags.push({
                    TagKey: key, 
                    TagValue: tags[key]
                });
            }

            await modifyResourceTags(modifyTagsRequest);
        } catch (e) {
            throw e;
        }
    }
}

module.exports = DeployFunction;
